package pgupdate

import groovy.text.SimpleTemplateEngine
import groovy.io.FileType
import groovy.lang.GroovyObject

import java.io.File

import org.apache.commons.io.*

public class UpdatePostgres {
	boolean continueProcessing = true
	def newline = System.getProperty("line.separator")
	def working_folder = System.getProperty("user.dir") + File.separator + "classes"
	
	//used to call UpdateLogger.groovy
	File UpdateLoggerClassFile = new File(working_folder, "UpdateLogger.groovy")
	Class UpdateLoggerClass = new GroovyClassLoader(getClass().getClassLoader()).parseClass(UpdateLoggerClassFile)
	GroovyObject updatelogger = (GroovyObject) UpdateLoggerClass.newInstance()

	/**
	 * Used to call into the custom logger class
	 * @return returns a new instance of the logger class
	 */
	def log(){
		return UpdateLoggerClass.newInstance()
	}
	
	/**
	 * Used to execute system commands. Will fail is the RC is not 0
	 * @param cmd | command we want to execute
	 * @return rc (return code of the executed command)
	 */
	def String launchCommandwEnv(cmd){
		if (continueProcessing){
	        log().debug("Running the following command: " + cmd)
			ProcessBuilder pb = new ProcessBuilder(cmd.split('  '))
			
			//set environment variables
			Map<String, String> env = pb.environment()
			env.put("LD_LIBRARY_PATH", System.properties['LD_LIBRARY_PATH'])
			env.put("LIBPATH", System.properties['LIBPATH'])
			env.put("PGUSER", System.properties['PGUSER'])
			env.put("PGPASSWORD", System.properties['PGPASSWORD'])
			env.put("PGPORT", System.properties['PGPORT'])
	
			pb.redirectErrorStream(true)
			//execute the command w/ parameters
			Process p = pb.start()
			//get output from the command
			InputStream stdout = p.getInputStream()
			BufferedReader reader = new BufferedReader (new InputStreamReader(stdout))
			def line = ""
			
			while ((line = reader.readLine()) != null){
				log().trace(line)
			}

			//print RC
			int rc = p.waitFor()
			log().debug("rc: " + rc)
			if(rc.toString() != "0"){
				log().error("The system has recieved a non-zero return code executing the following command and will not quit: " + cmd)
				System.exit(2)
			}
			return rc.toString()
		}
	}
	
	/**
	 * Used to execute system commands. Will  fail is the RC is not 0. NOTE: This is specific to the windows start up process. We do not try to read the output from the process because it will hang and never return.
	 * @param cmd | command we want to execute
	 * @return rc (return code of the executed command)
	 */
	def String launchCommandwEnvWin(cmd, cmdoutputfile){
		if (continueProcessing){
			log().debug("Running the following command: " + cmd)
			
			//piping the output to this text file in order to get around hanging issue with the bufferedReader
			cmd = cmd + "  >  " + cmdoutputfile
			def cmdlist = cmd.split('  ')
			cmdlist.plus(0, '/c')
			cmdlist.plus(0, 'cmd')
			ProcessBuilder pb = new ProcessBuilder(cmdlist)
			
			//set environment variables
			Map<String, String> env = pb.environment()
			env.put("LD_LIBRARY_PATH", System.properties['LD_LIBRARY_PATH'])
			env.put("LIBPATH", System.properties['LIBPATH'])
			env.put("PGUSER", System.properties['PGUSER'])
			env.put("PGPASSWORD", System.properties['PGPASSWORD'])
			env.put("PGPORT", System.properties['PGPORT'])
	
			pb.redirectErrorStream(true)
			//execute the command w/ parameters
			Process p = pb.start()
			
			//get return code
			int rc = p.waitFor()

			//read output back in from file
			File outputfile = new File(cmdoutputfile)
			String console_output = outputfile.text
			
			if (console_output.contains("Permission denied")){
				log().error(console_output)
			}
			else{
				log().trace(console_output)
			}

			//remove temp file
			outputfile.delete()
			outputfile.deleteOnExit()
			
			log().debug("rc: " + rc)
			if(rc.toString() != "0"){
				log().error("The system has recieved a non-zero return code executing the following command and will not quit: " + cmd)
				System.exit(2)
			}
			return rc.toString()
		}
	}
	/**
	 * Used to execute system commands when we don't care about the output (only the rc). Will NOT fail is the RC is not 0
	 * @param cmd | command we want to execute
	 * @return
	 */
	def String sendCommandwEnv(cmd){
		if (continueProcessing){
			log().debug("Running the following command: " + cmd)
			ProcessBuilder pb = new ProcessBuilder(cmd.split('  '))
			
			//set environment variables
			Map<String, String> env = pb.environment()
			env.put("LD_LIBRARY_PATH", System.properties['LD_LIBRARY_PATH'])
			env.put("LIBPATH", System.properties['LIBPATH'])
			env.put("PGUSER", System.properties['PGUSER'])
			env.put("PGPASSWORD", System.properties['PGPASSWORD'])
			env.put("PGPORT", System.properties['PGPORT'])
	
			pb.redirectErrorStream(true)
			//execute the command w/ parameters
			Process p = pb.start()

			InputStream stdout = p.getInputStream()
			BufferedReader reader = new BufferedReader (new InputStreamReader(stdout))
			def line = ""
			while ((line = reader.readLine()) != null){
				log().trace(line)
			}
			
			//print RC
			int rc = p.waitFor();
			log().debug("rc: " + rc)

			return rc.toString()
		}
	}
	/**
	 * Used to execute system commands when we care about the output
	 * @param cmd | command we want to execute
	 * @return
	 */
	
	def String sendCommandgetOutput(cmd){
		log().debug("Running the following command: " + cmd)
		ProcessBuilder pb = new ProcessBuilder(cmd.split('  '))
		
		Map<String, String> env = pb.environment()
		env.put("LD_LIBRARY_PATH", System.properties['LD_LIBRARY_PATH'])
		env.put("LIBPATH", System.properties['LIBPATH'])
		
		pb.redirectErrorStream(true)
		//execute the command w/ parameters
		Process p = pb.start()

		InputStream stdout = p.getInputStream()
		BufferedReader reader = new BufferedReader (new InputStreamReader(stdout))
		def line = ""
		def output = ""
		while ((line = reader.readLine()) != null){
			output = output + line
			log().trace(line)
		}
		int rc = p.waitFor();

		return output
		
	}

	/**
	 * Starts the specified server and verifies that it is running prior to returning
	 * @param script | location of the script we are starting
	 * @param binloc | location of the corresponding bin directory for verification purposes
	 * @return
	 */
	def startService(script, binloc){
		if (continueProcessing){
			log().info("Starting the service...")
			if(determineOS() == "WIN"){
				launchCommandwEnvWin(script +"  start", "startService_output.temp")
			}
			else{
				launchCommandwEnv(script +"  start")
			}
			
			//check if server is running
			String checkconnection = binloc + File.separator + "psql  -d  postgres  -c  select version()"
			def rc = sendCommandwEnv(checkconnection)
			def timeout = 6
			while(rc != "0"){
				//wait for 5 secs before trying to connect again
				log().debug("The server is not yet accepting connections. Will retry in 5 seconds...")
				sleep(5000)
				rc = sendCommandwEnv(checkconnection)
				timeout = timeout - 1

				if (timeout == 0)
				{
					log().error("The system reached a timeout while waiting for the server to come up. Processing will not continue for this dataserver.")
					continueProcessing = false
					break;
				}
			}
		}
	}
	
	/**
	 * Starts the specified server and verifies that it is running prior to returning
	 * @param script | location of the script we are stopping
	 * @return
	 */
	def stopService(script){
		if (continueProcessing){
			if(new File(script).exists()){
				log().info("Stopping the service...")
				launchCommandwEnv(script +"  stop")
			}
		}

	}
	

	/**
	 * Copies a file from one location to another
	 * @param source | file we are copying
	 * @param dest | location we are copying to
	 * @return
	 */
		
	def copyFile(source, dest){
		if (continueProcessing){
			log().debug("Attempting a file copy.")
			log().debug("Copying : " + source + " to " + dest)
			new File(dest) << new File(source).text
		}
	
	}
		

	/**
	 * Copies a file from one location to another while replacing 9.4 with 9.4_PREVIOUS as well as updating the permission on the resulting file. Definition also updates the servertype for windows
	 * @param source | file we are copying
	 * @param dest | location we are copying to
	 * @return
	 */
	def updateExecutionScript(script, target, datadir, datadirbck){
		if (continueProcessing){
			log().info("Updating" + target)
			File file = new File(script)
			File newfile = new File(target)
			String line
			//delete existing temp file if exists
			newfile.delete()
			
			def newline
			file.withReader { reader ->
				while ((line = reader.readLine())!=null) {

					if (line.contains("9.4")|| (line.contains("SERVERTYPE=services") && determineOS() == "WIN") || line.contains(datadir)) {
						newline = line.replace("9.4", "9.4_PREVIOUS").replace("SERVERTYPE=services", "SERVERTYPE=scripts").replace(datadir, datadirbck)
						
						if (determineOS() == "WIN") {
							newfile.append(newline + '\r\n')
						} else {
							newfile.append(newline + '\n')
						}
						
					} else {
						if (determineOS() == "WIN") {
							newfile.append(line + '\r\n')
						} else {
							newfile.append(line + '\n')
						}
					}
				}
			}
			//update permissions on unix
			if(determineOS() == "UNIX"){
				String chmodcom = "chmod  775  " + target
				launchCommandwEnv(chmodcom)
			}
		}
	}
	
	/**
	 * Creates a directory
	 * @param directory | directory location to be created
	 * @return
	 */
	
	def createDir(directory){
		if (continueProcessing){
			log().debug("Attempting to create the following directory: " + directory)
			new File(directory).mkdir()
			
			if(new File(directory).exists())
			{log().info("Successfully created the directory.")}
			else{
				log().error("Directory could not be created. Processing will not continue for this dataserver.")
				continueProcessing = false
			}
		}
	}
	
	/**
	 * Creates a backup of a specified file or folder. The file or folder backup will have the naming convention of <file>_backup_<version>
	 * @param directory | file or folder to be backed up
	 * @param version | version of the file or folder content
	 * @return the path/name of the backup
	 */

	def String renameToBackup(directory, version){
		if (continueProcessing){
			if(new File(directory).exists()){
			def backup = new File(directory+"_backup_" + version)
				if(backup.exists()){
					//if there is already a backup, delete it and recreate it.
					log().info("deleting existing backup")
					backup.deleteDir()
					if(backup.exists()){
						log().error("backup not deleted..")
					}
				}
			log().info("Changing the directory/file name to back up the following directory: " + directory)

			new File(directory).renameTo(backup)
			
			if(new File(directory+"_backup_" + version).exists())
				{log().debug("Successfully backed up the directory/file.")}
				else{
					log().error("Backup could not be created. Processing will not continue for this dataserver.")
					continueProcessing = false
				}
			}
			else{
				if(new File(directory+"_backup_" + version).exists())
					{log().info("Backup already created. Proceeding")}
					else{
						log().error("Unable to back up the following directory/file as it does not exist and processing will not continue for this dataserver: " + directory)
						continueProcessing = false
						}
			}
		}
		return directory + "_backup_" + version
	}
	
	
	def String copyToBackup(directory, sourcever, targetver){
		if (continueProcessing){
			log().info("The version for the backup creation is: " + sourcever)
			def source = new File(directory)
			def backup = new File(directory +"_backup_" + sourcever)
			def reentry = false
			
			if(source.exists()){
				if(backup.exists()){
					//if there is already a backup, delete it and recreate it. ONLY DO THIS IS THE CURRENT PG_VERSION OF THE DATADIR IS NOT AT THE TARGET VERSION
					if(getDatasrvVersiononDisk(directory).toString().minus(".").trim() != targetver){
						
						log().info("deleting existing backup")
						backup.deleteDir()
						
						if(new File(directory +"_backup_" + sourcever).exists()){
							log().error("backup not deleted..")
						}

					}
					else{
						log().info("Detected an existing backup and a data directory that is already at the target version. System will delete existing data directory and proceed with upgrade from the backup.")
						source.deleteDir()
						reentry = true
					}
				}
			
				if(!reentry){
					try{
		
						if(determineOS() == "WIN"){
							log().info("Renaming data directory: " + directory + " to : " + directory+"_backup_" + sourcever)
							def mvcmd = "cmd /c rename " + directory +  " " + directory+"_backup_" + sourcever
							mvcmd.execute()
							//still need to call the rename to change the system references. Even though we moved the file on disk it will still seem like the backup doesn't exist.
							source.renameTo(backup)
						}
						else
						{
							source.renameTo(backup)
						}
					}
					catch(IOException e){
						e.printStackTrace()
					}

			
					if(backup.exists())
					{
						log().info("Successfully backed up the directory/file.")
						//new File(directory).deleteDir()
						
						}
					else{
						log().error("Backup could not be created. Processing will not continue for this dataserver.")
						continueProcessing = false
					}
				}
			}
			else{
				//source.renameTo(backup)
				
				//the data directory is not found so check for a valid backup
				if(backup.exists())
					{log().warn("NOTE: System has detected no data directory but found a previously created backup. Proceeding")}
					else{
						log().error("Unable to back up the following directory/file as it does not exist and processing will not continue for this dataserver: " + directory)
						continueProcessing = false
						}
			}
		}
		return directory + "_backup_" + sourcever
	}


	/**
	 * Updates the specified file to contain version specific changes
	 * @param confFile | postgres.conf file to be updated
	 * @return
	 */
	
	def upgradePostgresConfFile(confFile, flatversion){
		if (continueProcessing){
			log().info("Attempting to update the postgres.conf file.")
			File file = new File(confFile)
			File newfile = new File(confFile + "_temp")
			String line
			
			file.withReader { reader ->
			   while ((line = reader.readLine())!=null) {
				   
				   if (line.contains("checkpoint_segments"))
				  {
					  def temp_list = line.split()
					  def c_s_value = temp_list[2].toInteger()
					  def max_wal_size = (3*c_s_value)*16
					  def min_wal_size = 80
					  
					  newfile.append('max_wal_size = ' + max_wal_size + 'GB #(3 * checkpoint_segments) * 16]MB gives equivalent value' + '\n' +  'min_wal_size = ' + min_wal_size + 'MB #80 MB default' + '\n')
				  }
				  else if (line.contains("unix_socket_directory"))
				  {
					  newfile.append(line.replace("unix_socket_directory", "unix_socket_directories") + '\n')
				  }
				  else
				  {
					  newfile.append(line + '\n')
				  }
	
			   }
			}
			
			//add additional settings
			newfile.append(getOptionfromConfFile(confFile + "_backup_" + flatversion , "dynamic_shared_memory_type") + '\n')
			
			//delete old file
			new File(confFile).delete()
			
			//rename new file
			new File(confFile + "_temp").renameTo(new File(confFile))
		}
	}

	/**
	 * grabs specified options from the specified file
	 * @param confFile | old postgres.conf file
	 * @param option | the specific option that we want to find
	 * @return
	 */
	def getOptionfromConfFile(confFile, option){
		File file = new File(confFile)
		String line
		String setting = ""
		file.withReader { reader ->
			while ((line = reader.readLine())!=null) {
				
				if (line.contains(option))
			   {

				   setting = line
				   log().debug("Located the following option: " + option + "... from the following file: " + confFile)
				  
			   }
 
			}
		 }
		 if(setting == ""){
			 log().debug("The following option: " + option + " was not found in the following file: " + confFile)
		 }
		 return setting
	}
	/**
	 * Dumps the database content to a file
	 * @param datasrvbin | pg_dumpall executable location
	 * @param dumpfile | the file to be created with the contents of the database
	 * @return
	 */
	
	def runPGDump(datasrvbin, dumpfile){
		if (continueProcessing){
			log().debug("Attempting to dump out Data Server contents for ")
			String dumpcmd =  datasrvbin + File.separator + "pg_dumpall  -f  " + dumpfile
			launchCommandwEnv(dumpcmd)
			
			if(new File(dumpfile).exists()){
				log().info("Successfully created database dump: " + dumpfile)
			}else
			{
				log().error("Unable to create dumpfile. Processing will not continue for this dataserver: " + dumpfile)
				continueProcessing = false
			}
		}
	}
	
	/**
	 * Updates the dump file to comment out commands related to the creation of the inituser
	 * This user is already created the initialization of the new db. Leaving these commands in there
	 * well cause warnings when the restore occurs
	 * @param dumpfile | dumpfile to be updated
	 * @param inituser | the db user that was used to create the dump
	 * @return
	 */
	
	def updateDumpFile(dumpfile, inituser){
		if (continueProcessing){
			log().info("Updating the dumpfile to not initate the super user")
			
			File file = new File(dumpfile)
			File newfile = new File(dumpfile + "_temp")
			String line
			
			def create_bol = false
			def alter_bol = false
			newfile.withWriter('UTF-8') { writer ->
				file.withReader('UTF-8') { reader ->
				   while ((line = reader.readLine())!=null) {
						   
						   if(!create_bol && !alter_bol){
							   if (line.contains("CREATE ROLE " + inituser))
							   {
								   writer << line.replace("CREATE ROLE", "-- CREATE ROLE") + newline
								   create_bol = true
							   }
							   else if (line.contains("ALTER ROLE " + inituser))
							   {
								  writer << line.replace("ALTER ROLE", "-- ALTER ROLE") + newline
								  alter_bol = true
							   }
							   else{
								 writer << line + newline
							   }
						   }
						   else{
							   writer << line + newline
						   }
				
						   
					   }
				  
	
			   }
			}
			
			if(newfile.exists()){
				//delete old files
				new File(dumpfile).delete()
				new File(dumpfile + "_temp").renameTo(new File(dumpfile))
				
				//change encoding to UTF8
				//new File(dumpfile).append("--", 'utf-8')
				
				//change encoding to UTF8
				//def f = new File(dumpfile + "_temp").getText()
				//new File(dumpfile + "_temp_utf8").write(f, 'utf-8')
				
				//delete old files
				//new File(dumpfile + "_temp").delete()
				
				//rename new file
				//new File(dumpfile + "_temp_utf8").renameTo(new File(dumpfile))
			}
			else{
				log().error("Unable to update dump file: " + dumpfile + ". Processing will not continue for this dataserver.")
				continueProcessing = false
			}
		 }
	}

	/**
	 * Updates the dump file to comment out commands related to the creation of the inituser
	 * This user is already created the initialization of the new db. Leaving these commands in there
	 * well cause warnings when the restore occurs
	 * @param dumpfile | dumpfile to be updated
	 * @param inituser | the db user that was used to create the dump
	 * @return
	 */
	
	def updateDump(dumpfile, inituser){
		if (continueProcessing){
			File file = new File(dumpfile)
			File newfile = new File(dumpfile + "_temp")
			
			def currentText = file.text
			def newText = currentText.replaceAll("CREATE ROLE " + inituser, "-- CREATE ROLE " + inituser)
			newText = newText.replaceAll("ALTER ROLE " + inituser, "-- ALTER ROLE " + inituser)
			
			newfile.write(newText, 'utf-8')
			
			if(newfile.exists()){
				//delete old file
				file.delete()
				
				//rename new file
				newfile.renameTo(file)
			}
			else
			{
				log().error("Unable to update dump file: " + dumpfile + ". Processing will not continue for this dataserver.")
				continueProcessing = false
			}
		}
	}
	/**
	 * 
	 * @param inituser | user being used to initialize the db
	 * @param datasrvbin | initdb bin location
	 * @param datasrvdata | data folder for the db
	 * @return
	 */
	
	def initializeNewDB(inituser, datasrvbin, datasrvdata){
		if (continueProcessing){
			log().debug("Attempting to initialize the database")
			String initcmd = datasrvbin + File.separator + "initdb" + "  -D  " + datasrvdata + "  -U  " + inituser + "  -A  trust  -E  UTF8  --locale=C"
			//String initcmd = datasrvbin + File.separator + "initdb" + "  -D  " + datasrvdata + "  -U  " + inituser + "  -A  trust  --locale=C"
			launchCommandwEnv(initcmd)
			
			if(new File(datasrvdata).exists()){
				log().info("Successfully able to initialize database.")
			}else
			{
				println("Unable to to initialize database. Processing will not continue for this dataserver.")
				continueProcessing = false
			}
		}
	}
	
	/**
	 * 
	 * @param datasrvbin
	 * @param inituser
	 * @param dumpFile
	 * @return
	 */
	
	def restoreDB(datasrvbin, inituser, dumpFile){
		if (continueProcessing){
			log().info("Attempting to restore Data Server contents for ")
			String restorecmd =  datasrvbin + File.separator + "psql  -d  postgres  -U  " + inituser +"  --set  ON_ERROR_STOP=1  -f  " + dumpFile
			launchCommandwEnv(restorecmd)
		}
	}
	
	/**
	 * Parses the database port from the corresponding script
	 * @param datasrvdir_script | postgres server script that contains the port
	 * @return
	 */
	
	def String findPortFromScript(datasrvdir_script){
		String port = "Unavailable"
		log().debug("Attempting to find the database port by scanning the following file: " + datasrvdir_script)
		File file = new File(datasrvdir_script)
		if(file.exists()){
			String line
			file.withReader { reader ->
			   while ((line = reader.readLine())!=null) {
				   
				   if (line.contains(" -p "))
				   {
					   
				  	def holder = line.split()
					//println("Found line: " + holder)
					holder.eachWithIndex {var, index ->
					//println(var)
					//println(index)
						if(var.equals("-p"))
						{
							port = holder[index.value + 1].minus('"')
							//println("The port has been discovered as: " + port)
						}
					  }
				   }
			   }
			}
		}
		else{
			//we will already know that the file doesn't exist from the precheck and will not do any actual work
		}
	
		return port
	}
	
	def String findDataDirFromScript(datasrvdir_script){
		String datadir = "Unavailable"
		if (continueProcessing){
			log().debug("Attempting to find the database data directory by scanning the following file: " + datasrvdir_script)
			File file = new File(datasrvdir_script)
			
			String line
			file.withReader { reader ->
			   while ((line = reader.readLine())!=null) {
				   
				   if (line.contains(" -D "))
				   {
					   
					  def holder = line.split()
					//println("Found line: " + holder)
					holder.eachWithIndex {var, index ->
					//println(var)
					//println(index)
						if(var.equals("-D"))
						{
							datadir = holder[index.value + 1].replaceAll('"',"")
							log().debug("The Data Directory has been discovered as: " + datadir)
						}
					  }
				   }
			   }
			}
		
			return datadir
		}
	}
	
	def String findConfigDirFromScript(datasrvdir_script){
		String configdir = "Unavailable"
		if (continueProcessing){
			log().debug("Attempting to find the database configuration directory by scanning the following file: " + datasrvdir_script)
			File file = new File(datasrvdir_script)
			
			String line
			file.withReader { reader ->
			   while ((line = reader.readLine())!=null) {
				   
				   if (line.contains("CONFIGDIR="))
				   {
					   
					  def holder = line.split("=")
					//println("Found line: " + holder)
					holder.eachWithIndex {var, index ->
					//println(var)
					//println(index)
						if(var.equals("CONFIGDIR"))
						{
							configdir = holder[index.value + 1].minus('"')
							log().debug("The configuration directory has been discovered as: " + configdir)
						}
					  }
				   }
			   }
			}
		
			return configdir
		}
	}
	
	/**
	 * Assumes that the data server execution script is named after the 12byte and grabs that 12byte/file name
	 * @param datasrvdir_script
	 * @return
	 */
	def String findTwelveByteFromScript(datasrvdir_script){
		def file = new File(datasrvdir_script)
		return file.name.replaceFirst(~/\.[^\.]+$/, '')
	}
	
	/**
	 * Determines if the system is Unix or Windows
	 * @return WIN or UNIX
	 */
	
	def String determineOS(){
		if (System.properties['os.name'].toLowerCase().contains('windows')) {
			return "WIN"
		} else {
			return "UNIX"
		}
	}
	
	/**
	 * Will determine to use shell or batch specification based on the current system OS
	 * @return "sh" or "bat"
	 */
	
	def String getShellorBat(){
		String os = determineOS()
		
		if (os.equalsIgnoreCase("WIN")){
			return ".bat"
		}
		else{
			return ".sh"
		}
		
	}
	
	/**
	 * Set environment variables
	 * @param user | postgres user
	 * @param pass | postgres password
	 * @param port | database port
	 * @return
	 */
	
	def setPGEnv(user, pass, port){
		System.properties['PGUSER'] = user
		System.properties['PGPASSWORD'] = pass
		System.properties['PGPORT'] = port
	}
	/**
	 * Set environment variables
	 * @param library | path to the library required
	 * @return
	 */
	
	def setLIBEnv(library){
		System.properties['LD_LIBRARY_PATH'] = library
		System.properties['LIBPATH'] = library
	}
	
	/**
	 * displays all the current system properties available top groovy
	 * @return
	 */
	def displaySystemProperties(){
		println("== Display system properties ==")
		System.properties.each { k,v->
			println ("$k = $v")
		}
	}
	
	/**
	 * Displays the current version of postgres as well as a list of all the schemas
	 * @param datasrvbin
	 * @return
	 */
	
	def displayVerification(datasrvbin){
		if (continueProcessing){
			String versioncmd = datasrvbin + File.separator + "psql  -d  postgres  -c  select version()"
			launchCommandwEnv(versioncmd)
			String listcmd = datasrvbin + File.separator + "psql  -d  postgres  -c  \\l"
			launchCommandwEnv(listcmd)
		}
	}
	
	/**
	 * Gets the PG_VERSION that is current on disk in the data directory
	 * @param datasrvdir_datadir | directory that contains the PG_VERSION file
	 * @return
	 */
	def String getDatasrvVersiononDisk(datasrvdir_datadir){
		def ver = 0
		def pg_verfile = new File(datasrvdir_datadir + File.separator + "PG_VERSION")
		if(pg_verfile.exists()){
			ver = new BigDecimal(pg_verfile.text.trim())
		}
		return ver
	}

	/**
	 * Disables security if it was configured on the system
	 * @param datasrvdir_datadir | data directory
	 */
	def disableSecurity(datasrvdir_datadir){
		def pgconf = new File(datasrvdir_datadir + File.separator + "postgresql.conf")
		def pghba = new File(datasrvdir_datadir + File.separator + "pg_hba.conf")
		
		def pgconf_temp = new File(datasrvdir_datadir + File.separator + "postgresql.conf.tmp")
		def pghba_temp = new File(datasrvdir_datadir + File.separator + "pg_hba.conf.tmp")
		
		if (continueProcessing){
			String line
			
			//handle postgresql.conf edits
			log().debug("Updating postgresl.conf to remove security settings.")
			pgconf.withReader { reader ->
				while ((line = reader.readLine())!=null) {
				   
				   if (line.contains("ssl_ciphers") || line.replaceAll("\\s", "").contains("ssl=on"))
				   {
				   	//ignore line
					 log().debug("Found security setting! Commenting it out...")
					 pgconf_temp.append("#" + line + newline )
				   }
				   else{
					 pgconf_temp.append(line + newline )
				   }
	 
				}
			}
			//delete old file
			pgconf.delete()
			//rename new file
			pgconf_temp.renameTo(pgconf)

			//handle certs
			log().debug("Removing any existing certs/keys")
			if( new File(datasrvdir_datadir + File.separator + "server.crt").exists() ){
				new File(datasrvdir_datadir + File.separator + "server.crt").delete()
			}
			if( new File(datasrvdir_datadir + File.separator + "server.key").exists() ){
				new File(datasrvdir_datadir + File.separator + "server.key").delete()
			}
			if( new File(datasrvdir_datadir + File.separator + "root.crt").exists() ){
				new File(datasrvdir_datadir + File.separator + "root.crt").delete()
			}
			if( new File(datasrvdir_datadir + File.separator + "root.crl").exists() ){
				new File(datasrvdir_datadir + File.separator + "root.crl").delete()
			}
		
		}
		
		//handle pg_hba.conf file
		log().debug("Updating pg_hba.conf file")
		def currentText = pghba.text
		def newText = currentText.replaceAll("clientcert=1" , "")
		newText = newText.replaceAll("hostssl", "host")
		newText = newText.replaceAll("cert", "md5")
		pghba_temp.write(newText)
		//delete old file
		pghba.delete()
		//rename new file
		pghba_temp.renameTo(pghba)

	}
	
	/**
	 * Checks to make sure the script has been provided valid system paths
	 */
	def checkSystemPaths(datasrvdir_home, datasrvdir_home_old, datasrvdir_config, datasrvdir_data, datasrvdir_bin, datasrvdir_bin_old, datasrvdir_lib_old, datasrvdir_conffile, datasrvdir_script){
		if(!new File(datasrvdir_home).exists()){
			log().error("Alert! The following directory/file does not exist and therefore processing will not continue for this dataserver will not continue. -->" + datasrvdir_home)
			continueProcessing = false
			}
		if(!new File(datasrvdir_home_old).exists()){
			log().error("Alert! The following directory/file does not exist and therefore processing will not continue for this dataserver will not continue. -->" + datasrvdir_home_old)
			continueProcessing = false
			}
		if(!new File(datasrvdir_config).exists()){
			log().error("Alert! The following directory/file does not exist and therefore processing will not continue for this dataserver will not continue. -->" + datasrvdir_config)
			continueProcessing = false
			}
		if(!new File(datasrvdir_data).exists()){
			//data directory does not exist so check for a valid backup directory (IMPORTANT FOR REENTRY: IN CASE SCRIPT EXITED DURING EXECUTION AND THE BACKUP WAS ALREADY CREATED)
			if(!new File(datasrvdir_data + "_backup_" + getVersion(datasrvdir_home_old)))
				{
					log().error("Alert! The following directory/file does not exist and therefore processing will not continue for this dataserver will not continue. -->" + datasrvdir_data)
					log().error("Alert! The system also could not find a valid backup directory.")
					continueProcessing = false
				}
			}
		if(!new File(datasrvdir_bin).exists()){
			log().error("Alert! The following directory/file does not exist and therefore processing will not continue for this dataserver will not continue. -->" + datasrvdir_bin)
			continueProcessing = false
			}
		if(!new File(datasrvdir_bin_old).exists()){
			log().error("Alert! The following directory/file does not exist and therefore processing will not continue for this dataserver will not continue. -->" + datasrvdir_bin_old)
			continueProcessing = false
			}
		//if(!new File(datasrvdir_conffile).exists()){
		//	println("Alert! The following directory/file does not exist and therefore processing will not continue for this dataserver will not continue. -->" + datasrvdir_conffile)
		//	continueProcessing = false
		//	}
		if(!new File(datasrvdir_script).exists()){
			log().error("Alert! The following directory/file does not exist and therefore processing will not continue for this dataserver will not continue. -->" + datasrvdir_script)
			continueProcessing = false
			}

	}
	
	/**
	 * Gets postgres version from installed binaries
	 * @param datasrvdir_bin | binary bin location
	 * @param datasrvdir_bin | binary library location
	 */
	def String getVersion(datasrvdir_home){
		log().debug("Getting the version from the postgres binary location.")
		def datasrvdir_lib = datasrvdir_home + File.separator +"lib"
		def datasrvdir_bin = datasrvdir_home + File.separator +"bin"
		
		//Need to make sure the library is set prior to any call to the binary location
		setLIBEnv(datasrvdir_lib)
		
		String ver = "0"
		String execmd = ""
		if(new File(datasrvdir_bin).exists()){
			
			if(determineOS() == "WIN"){
				execmd = datasrvdir_bin + File.separator + "postgres.exe  -V"
			}
			else{
				execmd = datasrvdir_bin + File.separator + "postgres  -V"
			}
			//get rid of the end of the minor version
			ver = sendCommandgetOutput(execmd).minus("postgres (PostgreSQL)")
			ver = ver.substring(0, 4)
		}
		else
		{
			log().error("Postges binary location is not available! -> " + datasrvdir_bin)
			System.exit(1)
		}
		return ver
	}

	/**
	 * Upgrades a Postgres instance from 9.1 to 9.5 or 9.4 to 9.5
	 * @param twelvebyte | Dataserver 12byte
	 * @param home | SASHome location of the 9.5 binaries
	 * @param config | configured location
	 * @param user | pg user
	 * @param pass | pg pass
	 * @return
	 */
	def int updateInstance(twelvebyte, home, config, data, user, pass){
		continueProcessing = true
		log().info("Starting Postgres Update for: " + twelvebyte)
		String datasrvdir_twelvebyte = twelvebyte.toLowerCase()
		String datasrvdir_home = home
		String datasrvdir_config = config
		String datasrvdir_data = data
		String datasrvdir_os = determineOS()
		String datasrvdir_home_old = datasrvdir_home.minus("9.4").plus("9.4_PREVIOUS")
		String datasrvdir_script = datasrvdir_config + File.separator + datasrvdir_twelvebyte + getShellorBat()
		String datasrvdir_temp = datasrvdir_config + File.separator + "temp"
		String datasrvdir_temp_script_name = datasrvdir_twelvebyte + "_temp" + getShellorBat()
		String datasrvdir_temp_script = datasrvdir_config + File.separator + datasrvdir_temp_script_name
		String datasrvdir_bin = datasrvdir_home + File.separator + "bin"
		String datasrvdir_bin_old = datasrvdir_home_old + File.separator + "bin"
		String datasrvdir_lib = datasrvdir_home + File.separator + "lib"
		String datasrvdir_lib_old = datasrvdir_home_old + File.separator + "lib"
		String datasrvdir_conffile = datasrvdir_data + File.separator + "postgresql.conf"
		String datasrvdir_dumpfile = datasrvdir_temp + File.separator + "db_update.dat"
		
		//check to make sure all the paths are valid before doing anything
		checkSystemPaths(datasrvdir_home, datasrvdir_home_old, datasrvdir_config, datasrvdir_data, datasrvdir_bin, datasrvdir_bin_old, datasrvdir_lib_old, datasrvdir_conffile, datasrvdir_script)
		
		//set env variables
		setPGEnv(user, pass, findPortFromScript(datasrvdir_script))//get the port from a file on disk
		setLIBEnv(datasrvdir_home_old + File.separator + "lib")

		//get db user
		String datasrvdir_inituser = System.properties['PGUSER']
		
		//set SOURCE AND TARGET VERSIONS
		String currentver = getVersion(datasrvdir_home_old)//SCOTT!!get from the binaries in SASHOME!
		String targetver = getVersion(datasrvdir_home) //SCOTT!!get from the binaries in SASHOME!
		String verforfile_current = currentver.toString().minus(".").trim()
		String verforfile_target = targetver.toString().minus(".").trim()
		String datasrvdir_conffile_old = datasrvdir_data + "_backup_" + verforfile_current + File.separator + "postgresql.conf"
		
		
		log().trace("currentver: " + currentver)
		log().trace("targetver: " + targetver)
		log().trace("verforfile_current: " + verforfile_current)
		log().trace("verforfile_target: " + verforfile_target)
		
		//display current system properties
		//displaySystemProperties()
		
		//print important values for debugging purposes
		log().debug("Data Server Home Directory is " + datasrvdir_home)
		log().debug("Data Server Config Directory is " + datasrvdir_config)
		log().debug("Data Server Data Directory is " + datasrvdir_data)
		log().debug("Data Server Shell Script is " + datasrvdir_script)
		log().debug("Data Server Temp Location is " + datasrvdir_temp)
		log().debug("Data Server Temp Script Location is " + datasrvdir_temp_script)
		log().debug("Data Server Init User is " + datasrvdir_inituser)
		
		
		
		
		//1. Check to make sure the database is not already at Postgres version 9.5
		
		//stop servers from that may be running from previous attempt
		stopService(datasrvdir_temp_script)
		stopService(datasrvdir_script)
		
		//2. Rename to backup the data dir of the postgres instance (needs to be retry friendly in the event of failure at any stage for example due to running out of disk space).
		String datasrvdir_data_bck = copyToBackup(datasrvdir_data, verforfile_current, verforfile_target)

		//3. Copy the webinfdsvr sh/bat from existing config location and change the SASHOME reference to point to the old content; now name 9.4_PREVIOUS (for Windows make sure SERVERTYPE=scripts).
		createDir(datasrvdir_temp)
		updateExecutionScript(datasrvdir_script, datasrvdir_temp_script, datasrvdir_data, datasrvdir_data_bck)
		
		//4. Run the script to start the service and wait for the port to come up (should use same port number per that instance configuration)
		startService(datasrvdir_temp_script, datasrvdir_bin_old)
		displayVerification(datasrvdir_bin_old)
		
		//5. Use pg_dumpall to dump all databases in data server to sql file
		//Modify the dump to comment out references to the database instance user (example see webinfdsvrcTasks.xml: under <!-- comment out recreating the superuser from the dump -->)
		runPGDump(datasrvdir_bin_old, datasrvdir_dumpfile)
		updateDumpFile(datasrvdir_dumpfile, datasrvdir_inituser)
	 
		//6. Stop the service and wait for the port to be available.
		stopService(datasrvdir_temp_script)

		//7. Run the initdb command from Postgres version 9.5 folder
		setLIBEnv(datasrvdir_home + File.separator + "lib")
		initializeNewDB(datasrvdir_inituser, datasrvdir_bin, datasrvdir_data)
		
		//8. Backup the postgresql.conf file from version 9.5 and copy in the conf file from version 9.1.
		renameToBackup(datasrvdir_conffile, verforfile_target)
		copyFile(datasrvdir_conffile_old, datasrvdir_conffile)
		
		//8a. Modify the postgresql.conf file with the additions/updates needed for postgres 9.5
		upgradePostgresConfFile(datasrvdir_conffile, verforfile_target)
		
		//8b. Disable security is it was enabled
		disableSecurity(datasrvdir_data)
		
		//9. Start the service using the script in the data server instance Config Lev dir and wait for the port to be available.
		startService(datasrvdir_script, datasrvdir_bin)
		
		//10. Use psql to restore the sql file dump
		restoreDB(datasrvdir_bin, datasrvdir_inituser, datasrvdir_dumpfile)
		
		//11. Display information to verify the dump restored successfully (Postgres version and list of databases)
		displayVerification(datasrvdir_bin)
		
		//12. Stop the service
		stopService(datasrvdir_script)

		//13. Copy over old pg_hba.conf from backup data directory
		renameToBackup(datasrvdir_data + File.separator + "pg_hba.conf", verforfile_target)
		copyFile(datasrvdir_data + "_backup_" + verforfile_current + File.separator + "pg_hba.conf", datasrvdir_data + File.separator + "pg_hba.conf")
		
		//13a.call this again to remove security setting that may have been added back via the pg_hba.conf copy from backup
		disableSecurity(datasrvdir_data)
		
		//14. Clean Up
		if (continueProcessing){
			//remove backup if flag is set
			if(System.properties['REMOVEDATABACKUP'] == true){
				log().info("The --REMOVEDATABACKUP option has been set. Deleting the following folder: " + datasrvdir_data + "_backup_" + verforfile_current)
				new File(datasrvdir_data + "_backup_" + verforfile_current).deleteDir()
			}
			//keep dumpfile if flag is set
			if(System.properties['KEEPDUMP'] != true){
				log().info("The --KEEPDUMP option has NOT been set. The following dumpfile will be deleted: " + datasrvdir_dumpfile)
				new File(datasrvdir_dumpfile).delete()
				new File(datasrvdir_temp).deleteDir()

			}
			else{
				log().info("The --KEEPDUMP option has been set. The following dumpfile will NOT be deleted: " + datasrvdir_dumpfile)
			}
			
			new File(datasrvdir_temp_script).delete()
			//new File(datasrvdir_conffile + "_backup_" + verforfile_target).delete()
			log().info("Postgres Update for: " + twelvebyte + " complete.")
			
			return 0
		}
		else{
			log().error("ERROR: Postgres Update for: " + twelvebyte + " NOT complete.")
			
			return -2 
		}
	}
	
	
	static void main(def args){

	}
}
